@@ -1,7 +1,7 @@
 package Complete::Bash;
 
-our $DATE = '2019-06-25'; # DATE
-our $VERSION = '0.321'; # VERSION
+our $DATE = '2019-06-26'; # DATE
+our $VERSION = '0.322'; # VERSION
 
 use 5.010001;
 use strict;
@@ -397,6 +397,16 @@ sub join_wordbreak_words {
     [$new_words, $cword];
 }
 
+sub _columns {
+    # XXX need to cache?
+    if (eval { require Term::Size; 1 }) {
+        my ($columns, undef) = Term::Size::chars();
+        $columns;
+    } else {
+        $ENV{COLUMNS} // 80;
+    }
+}
+
 $SPEC{format_completion} = {
     v => 1.1,
     summary => 'Format completion for output (for shell)',
@@ -514,24 +524,45 @@ sub format_completion {
     $opts //= {};
 
     $hcomp = {words=>$hcomp} unless ref($hcomp) eq 'HASH';
-    my $comp     = $hcomp->{words};
+    my $words    = $hcomp->{words};
     my $as       = $hcomp->{as} // 'string';
     # 'escmode' key is deprecated (Complete 0.11-) and will be removed later
     my $esc_mode = $hcomp->{esc_mode} // $hcomp->{escmode} // 'default';
     my $path_sep = $hcomp->{path_sep};
 
-    if (defined($path_sep) && @$comp == 1) {
+    my @res;
+    my @summaries;
+    my $has_summary;
+
+  FORMAT_MESSAGE:
+    # display a message instead of list of words. we send " " (ASCII space)
+    # which bash does not display, so we can display a line of message while the
+    # user does not get the message as the completion. I've also tried \000 to
+    # \037 instead of space (\040) but nothing works better.
+    if (defined $hcomp->{message}) {
+        my $msg = $hcomp->{message};
+        if ($msg =~ /\A /) {
+            $msg =~ s/\A +//;
+            $msg = " (empty message)" unless length $msg;
+        }
+        @res = (sprintf("%-"._columns()."s", $msg), " ");
+        goto RETURN_RES;
+    }
+
+  WORKAROUND_PREVENT_BASH_FROM_INSERTING_SPACE:
+    if (defined($path_sep) && @$words == 1) {
         my $re = qr/\Q$path_sep\E\z/;
         my $word;
-        if (ref($comp->[0]) eq 'HASH') {
-            $comp = [$comp->[0], {word=>"$comp->[0] "}] if
-                $comp->[0]{word} =~ $re;
+        if (ref($words->[0]) eq 'HASH') {
+            $words = [$words->[0], {word=>"$words->[0] "}] if
+                $words->[0]{word} =~ $re;
         } else {
-            $comp = [$comp->[0], "$comp->[0] "]
-                if $comp->[0] =~ $re;
+            $words = [$words->[0], "$words->[0] "]
+                if $words->[0] =~ $re;
         }
     }
 
+  WORKAROUND_WITH_WORDBREAKS:
     # this is a workaround. since bash breaks words using characters in
     # $COMP_WORDBREAKS, which by default is "'@><=;|&(: this presents a problem
     # we often encounter: if we want to provide with a list of strings
@@ -546,7 +577,7 @@ sub format_completion {
     if (defined($opts->{word})) {
         if ($opts->{word} =~ s/(.+[\@><=;|&\(:])//) {
             my $prefix = $1;
-            for (@$comp) {
+            for (@$words) {
                 if (ref($_) eq 'HASH') {
                     $_->{word} =~ s/\A\Q$prefix\E//i;
                 } else {
@@ -556,10 +587,8 @@ sub format_completion {
         }
     }
 
-    my @res;
-    my @summaries;
-    my $has_summary;
-    for my $entry (@$comp) {
+  ESCAPE_WORDS:
+    for my $entry (@$words) {
         my $word    = ref($entry) eq 'HASH' ? $entry->{word}    : $entry;
         my $summary = (ref($entry) eq 'HASH' ? $entry->{summary} : undef) // '';
         if ($esc_mode eq 'shellvar') {
@@ -581,14 +610,7 @@ sub format_completion {
         last unless $has_summary;
         last unless $opts->{show_summaries} //
             $ENV{COMPLETE_BASH_SHOW_SUMMARIES} // 1;
-        my $columns = do {
-            if (eval { require Term::Size; 1 }) {
-                my ($columns, undef) = Term::Size::chars();
-                $columns;
-            } else {
-                $ENV{COLUMNS} // 80;
-            }
-        };
+        my $columns = _columns();
         my $width = 8;
         for (@res) {
             $width = length if $width < length;
@@ -603,6 +625,7 @@ sub format_completion {
         }
     } # INSERT_SUMMARIES
 
+  RETURN_RES:
     if ($as eq 'array') {
         return \@res;
     } else {
@@ -625,7 +648,7 @@ Complete::Bash - Completion routines for bash shell
 
 =head1 VERSION
 
-This document describes version 0.321 of Complete::Bash (from Perl distribution Complete-Bash), released on 2019-06-25.
+This document describes version 0.322 of Complete::Bash (from Perl distribution Complete-Bash), released on 2019-06-26.
 
 =head1 DESCRIPTION
 
